class AttributeCollection:
    def __init__(self, client, corpus):
        self.client = client
        self.corpus = corpus
        self.alignment = AlignmentAttributeCollection(client, corpus)
        self.positional = PositionalAttributeCollection(client, corpus)
        self.structural = StructuralAttributeCollection(client, corpus)


class Attribute:
    def __init__(self, client, corpus, name):
        self.client = client
        self.corpus = corpus
        self.name = name
        self._name = '{}.{}'.format(corpus.name, name)
        self.size = client.api.cl_attribute_size(self._name)

    def drop(self):
        return self.client.api.cl_drop_attribute(self._name)


class AlignmentAttributeCollection:
    def __init__(self, client, corpus):
        self.client = client
        self.corpus = corpus

    def get(self, name):
        return AlignmentAttribute(self.client, self.corpus, name)

    def list(self):
        return [AlignmentAttribute(self.client, self.corpus, attr) for attr in
                self.client.api.corpus_alignment_attributes(self.corpus.name)]


class AlignmentAttribute(Attribute):
    def cpos_by_ids(self, id_list):
        return self.client.api.cl_alg2cpos(self._name, id_list)

    def ids_by_cpos(self, cpos_list):
        return self.client.api.cl_cpos2alg(self._name, cpos_list)


class PositionalAttributeCollection:
    def __init__(self, client, corpus):
        self.client = client
        self.corpus = corpus

    def get(self, name):
        return PositionalAttribute(self.client, self.corpus, name)

    def list(self):
        return [PositionalAttribute(self.client, self.corpus, attr) for attr in
                self.client.api.corpus_positional_attributes(self.corpus.name)]


class PositionalAttribute(Attribute):
    def __init__(self, client, corpus, name):
        super(PositionalAttribute, self).__init__(client, corpus, name)
        self.lexicon_size = client.api.cl_lexicon_size(self._name)

    def cpos_by_id(self, id):
        return self.client.api.cl_id2cpos(self._name, id)

    def cpos_by_ids(self, id_list):
        return self.client.api.cl_idlist2cpos(self._name, id_list)

    def freqs_by_ids(self, id_list):
        return self.client.api.cl_id2freq(self._name, id_list)

    def ids_by_cpos(self, cpos_list):
        return self.client.api.cl_cpos2id(self._name, cpos_list)

    def ids_by_regex(self, regex):
        return self.client.api.cl_regex2id(self._name, regex)

    def ids_by_values(self, value_list):
        return self.client.api.cl_str2id(self._name, value_list)

    def values_by_cpos(self, cpos_list):
        return self.client.api.cl_cpos2str(self._name, cpos_list)

    def values_by_ids(self, id_list):
        return self.client.api.cl_id2str(self._name, id_list)


class StructuralAttributeCollection:
    def __init__(self, client, corpus):
        self.client = client
        self.corpus = corpus

    def get(self, name):
        return StructuralAttribute(self.client, self.corpus, name)

    def list(self, filters={}):
        attrs = [StructuralAttribute(self.client, self.corpus, attr) for attr
                 in self.client.api.corpus_structural_attributes(
                    self.corpus.name)]
        for k, v in filters.items():
            if k == 'part_of':
                attrs = list(filter(lambda x: x.name.startswith(v.name + '_'),
                                    attrs))
        return attrs


class StructuralAttribute(Attribute):
    def __init__(self, client, corpus, name):
        super(StructuralAttribute, self).__init__(client, corpus, name)
        self.has_values = \
            client.api.corpus_structural_attribute_has_values(self._name)

    def cpos_by_id(self, id):
        return self.client.api.cl_struc2cpos(self._name, id)

    def ids_by_cpos(self, cpos_list):
        return self.client.api.cl_cpos2struc(self._name, cpos_list)

    def lbound_by_cpos(self, cpos_list):
        return self.client.api.cl_cpos2lbound(self._name, cpos_list)

    def rbound_by_cpos(self, cpos_list):
        return self.client.api.cl_cpos2rbound(self._name, cpos_list)

    def values_by_ids(self, id_list):
        return self.client.api.cl_struc2str(self._name, id_list)

    def export(self, first_cpos, last_cpos, context=0, expand_lists=False):
        first_id, last_id = self.ids_by_cpos([first_cpos, last_cpos])
        c = (first_cpos, last_cpos)
        lc = rc = None
        if context == 0:
            cpos_list = list(range(first_cpos, (last_cpos + 1)))
        else:
            lc_lbound = self.cpos_by_id(max(0, (first_id - context)))[0]
            if lc_lbound != first_cpos:
                lc_rbound = max(0, (first_cpos - 1))
                lc = (lc_lbound, lc_rbound)
                cpos_list_lbound = lc_lbound
            else:
                cpos_list_lbound = first_cpos
            rc_rbound = \
                self.cpos_by_id(min((last_id + context), (self.size - 1)))[1]
            if rc_rbound != last_cpos:
                rc_lbound = min((last_cpos + 1), (self.corpus.size - 1))
                rc = (rc_lbound, rc_rbound)
                cpos_list_rbound = rc_rbound
            else:
                cpos_list_rbound = last_cpos
            cpos_list = list(range(cpos_list_lbound, (cpos_list_rbound + 1)))
        if expand_lists:
            match = {'lc': list(range(lc[0], (lc[1] + 1))),
                     'c': list(range(c[0], (c[1] + 1))),
                     'rc': list(range(rc[0], (rc[1] + 1)))}
        else:
            match = {'lc': lc, 'c': c, 'rc': rc}
        lookups = self.corpus.lookups_by_cpos(cpos_list)
        return {'match': match, **lookups}
